SENTRY_DSN = ''
MMSTATS_BASE_DIR = ''

SERVICES = {
    'email_sender': {
        'queues': {
            'email_sender__register': {
                'durable': True,
                'want_reply': False,
                'x-message-ttl': 600 * 1000,
            },
            'email_sender__password_reset': {
                'durable': True,
                'want_reply': False,
                'x-message-ttl': 600 * 1000,
            },
            'email_sender__pkgups_notif': {
                'durable': True,
                'want_reply': False,
                'x-message-ttl': 3600 * 1000,
            },
            'email_sender__contact_form': {
                'durable': True,
                'want_reply': False,
                'x-message-ttl': 36000 * 1000,
            },
        },
    },

    'nut_indexing': {
        'queues': {
            'nut_indexing': {
                'durable': False,
                'want_reply': True,
                'x-message-ttl': 30 * 1000,
                'default_timeout': 5.0,
            },
        },
    },

###

    'example': {
        'queues': {
            'example__q1': {
                'durable': False,
                'want_reply': True,
            },
            'example__q2': {
                'durable': False,
                'want_reply': True,
            },
        },
    },
}




import datetime
import os
import os.path
import socket
import time

import amqplib.client_0_8 as amqp

from amqpservices.base import log, config, AMQPBase, LOCALHOST_CONNECTION
from amqpservices.serializers.nop import NopSerializer

try:
    import mmstats

    class AMQPServicesStats(mmstats.MmStats):
        worker_id = mmstats.StringField(label='amqpservices.worker_id')
        processed = mmstats.CounterField(label='amqpservices.processed')
        exceptions = mmstats.CounterField(label='amqpservices.exceptions')
        last_before_callback = mmstats.DoubleField(label='amqpservices.last_before_callback')
        last_after_callback = mmstats.DoubleField(label='amqpservices.last_after_callback')
        state = mmstats.StringField(label='amqpservices.state')
        totaltime_callback = mmstats.DoubleField(label='amqpservices.totaltime_callback')
        avg_callback = mmstats.AverageField(label='amqpservices.avg_callback')

    USE_MMSTATS = True
    log.info('Using mmstats')

except ImportError:
    USE_MMSTATS = False
    log.info('Not using mmstats - no such module')


def worker_id(service):
    process_created = datetime.datetime.fromtimestamp(os.path.getmtime('/proc/self'))
    return '{}-{}-{}-{}'.format(
            service,
            socket.gethostname(),
            process_created.strftime('%y%m%d.%H%M%S'),
            os.getpid())

class Server(AMQPBase):
    '''
    :param service: name of a service from the config module
    :param callback: a callable supporting a call ``callback(props, body)``, where ``props`` is a dictionary of AMQP message properties and ``body`` is a message body - either a raw string or a rich type produced by a serializer's ``from_bytes`` method.
    :param serializer: an object implementing `~amqpservices.serializers.base.Serializer` methods used for de-serializing requests and serializing responses from the ``callback``
    :param connparams: parameters for a connection to a AMQP broker

    '''
    def __init__(self, service, callback, serializer=None,
            connparams=LOCALHOST_CONNECTION):
        self.init_connection_and_channel(connparams, connect_timeout=5)
        try:
            self.callback = callback
            self.service = service
            self.serializer = serializer or NopSerializer()

            self.worker_id = worker_id(service)

            self.stats = None
            if USE_MMSTATS:
                path = getattr(config.module, 'MMSTATS_BASE_DIR', '')
                if path:
                    self.stats = AMQPServicesStats(path=path, filename=self.worker_id + '.stats')
                    self.stats.worker_id = self.worker_id
                else:
                    log.warn('Empty or unset MMSTATS_BASE_DIR config variable, stats not available')

            #for sig in signal.SIGHUP, signal.SIGUSR1, signal.SIGUSR2:
            #    signal.signal(sig, self._signal_catched)

            log.info('Starting setup of worker pid=%d name=%s', os.getpid(), self.worker_id)
            for queue_name, queue_config in config.module.SERVICES[self.service]['queues'].items():
                self._create_public_queue(queue_name, queue_config)
                self._register_service_request_consumer(queue_name, self.callback)
            self._create_bc_queue()
            self._register_service_request_consumer(self.bc_queue_name, self.callback)
        except:
            log.exception('Server.__init__ failed')
            config.sentry.captureException()
            raise

    def _create_public_queue(self, queue_name, queue_config):
        arguments = {}
        if queue_config.get('x-message-ttl') is not None:
            arguments['x-message-ttl'] = queue_config['x-message-ttl']
        self.ch.queue_declare(queue=queue_name,
                durable=queue_config['durable'],
                auto_delete=queue_config.get('auto_delete', False),
                arguments=arguments,
                )
        # Routing keys are the same as queue names
        self.ch.queue_bind(exchange='amq.direct', queue=queue_name, routing_key=queue_name)
        log.info('Created public service queue %s' % queue_name)

    def _bc_routing_key(self):
        return 'broadcast.' + self.service

    def _create_bc_queue(self):
        self.bc_queue_name, _, _ = self.ch.queue_declare(exclusive=False)
        self.ch.queue_bind(exchange='amq.direct', queue=self.bc_queue_name, routing_key=self._bc_routing_key())

    def _register_service_request_consumer(self, queue_name, callback):
        def _request_callback(msg):
            log.info('Got msg')
            log.debug('Msg props, body: <%s> <%s>', msg.properties, msg.body[:1000])
            if self.stats:
                self.stats.state = 'callback'
                self.stats.last_before_callback = time.time()
            before_callback = time.time()
            error_condition = False
            try:
                body_rich = self.serializer.from_bytes(msg.body)
                ret_rich = callback(msg.properties, body_rich)
                ret = self.serializer.to_bytes(ret_rich)
                assert ret is None or isinstance(ret, str), \
                    'Serializer must serialize callback\'s response to None or str'
            except:
                log.exception('Exception thrown by the callback')
                config.sentry.captureException()
                if self.stats:
                    self.stats.exceptions.inc()
                # Reject the message. We are sleeping 1 second here to avoid
                # high CPU/disk usage caused by constant ping-pong between
                # our process that rejects the message and RabbitMQ which
                # redelivers the message to us. A small cost is that we
                # block this consumer for a second.
                log.warn('Sleeping before rejection')
                time.sleep(1)
                msg.channel.basic_reject(msg.delivery_tag, True)
                log.warn('Rejected')
                error_condition = True
            else:
                if self.stats:
                    self.stats.processed.inc()
                log.debug('Response callback finished with success')

            after_callback = time.time()
            spent_callback = after_callback - before_callback
            if self.stats:
                self.stats.totaltime_callback += spent_callback
                self.stats.avg_callback.add(spent_callback)
                self.stats.last_after_callback = time.time()

            if not error_condition and 'reply_to' in msg.properties:
                if self.stats:
                    self.stats.state = 'response'
                body = ret if ret is not None else '<None>'
                msg.channel.basic_publish(amqp.basic_message.Message(
                    body=body,
                    correlation_id=msg.properties['correlation_id']),
                    exchange='amq.direct',
                    routing_key=msg.properties['reply_to'])
                log.info('Processed msg')
                log.debug('Sent reply with body="%s" to %s', body, msg.properties['reply_to'])
            if not error_condition:
                if self.stats:
                    self.stats.state = 'ack'
                msg.channel.basic_ack(msg.delivery_tag)
            if self.stats:
                self.stats.state = 'wait'
        self.ch.basic_consume(queue=queue_name, callback=_request_callback)

    def loop(self):
        '''Infinite server loop: reads client's requests and sends responses
        using a callback passed in the ``__init__`` method.
        '''

        try:
            while self.ch.callbacks:
                self._process_events(timeout=None)
        except:
            log.exception('')
            log.critical('Unhandled exception from loop()')
            config.sentry.captureException()
            raise

    def _signal_catched(self, signo, frame):
        log.warn('Not handling signal %s', signo)




# Module name clash - "json" is the current module as well as
# a stdlib module. Use importlib to force importing
# the latter.
import importlib
json = importlib.import_module('json', '')

from amqpservices.serializers import base
from amqpservices import Server, Client
from amqpservices.base import LOCALHOST_CONNECTION


class JSONSerializer(base.Serializer):

    def from_bytes(self, bs):
        return json.loads(bs)

    def to_bytes(self, p):
        return json.dumps(p)


def expose(f):
    f.amqpservices_exposed = True
    return f

def is_exposed(f):
    return getattr(f, 'amqpservices_exposed', False)


class JSONDispatcher(base.ResponseProducer):

    def __init__(self, impl):
        self.impl = impl

    def __call__(self, properties, json_request):
        method_name, args, kwargs = json_request['method'], \
            json_request.get('args', []), \
            json_request.get('kwargs', {})
        method = getattr(self.impl, method_name)
        if not is_exposed(method):
            raise AttributeError('Method %s is not exposed' % method_name)
        result = method(*args, **kwargs)
        return result


class JSONFromListDispatcher(base.ResponseProducer):

    def __init__(self, method_list):
        self.method_list = method_list
        self._name_to_method = {m.__name__: m for m in self.method_list}

    def __call__(self, properties, json_request):
        method_name, args, kwargs = json_request['method'], \
            json_request.get('args', []), \
            json_request.get('kwargs', {})
        if method_name not in self._name_to_method:
            raise AttributeError('Method not found: %s' % method_name)
        result = self._name_to_method[method_name](*args, **kwargs)
        return result


class _MethodProxy(object):

    def __init__(self, proxy, name):
        self.proxy = proxy
        self.name = name

    def __call__(self, *args, **kwargs):
        request = {
                'method': self.name,
                'args': args,
                'kwargs': kwargs,
                }
        return self.proxy.client.send_request(self.proxy.routing_key, request)


class JSONCallerProxy(object):

    def __init__(self, client, routing_key):
        self.client = client
        self.routing_key = routing_key

    def __getattr__(self, name):
        return _MethodProxy(self, name)


def json_server(service, impl):
    return Server(service=service,
            serializer=JSONSerializer(),
            callback=JSONDispatcher(impl))

def json_methods_server(service, fun_list):
    return Server(service=service,
            serializer=JSONSerializer(),
            callback=JSONFromListDispatcher(fun_list))

def json_proxy(routing_key, connparams=LOCALHOST_CONNECTION):
    client = Client(connparams=connparams, serializer=JSONSerializer())
    return JSONCallerProxy(client, routing_key)




import socket

import amqplib.client_0_8 as amqp
from amqplib.client_0_8 import transport
from amqplib.client_0_8.exceptions import AMQPConnectionException
from amqplib.client_0_8.exceptions import AMQPChannelException


### Code from kombu BEGIN

transport.AMQP_PROTOCOL_HEADER = 'AMQP\x01\x01\x08\x00'

class Connection(amqp.Connection):  # pragma: no cover

    def _do_close(self, *args, **kwargs):
        # amqplib does not ignore socket errors when connection
        # is closed on the remote end.
        try:
            super(Connection, self)._do_close(*args, **kwargs)
        except socket.error:
            pass

    def _dispatch_basic_return(self, channel, args, msg):
        reply_code = args.read_short()
        reply_text = args.read_shortstr()
        exchange = args.read_shortstr()
        routing_key = args.read_shortstr()

        exc = AMQPChannelException(reply_code, reply_text, (50, 60))
        if channel.events["basic_return"]:
            for callback in channel.events["basic_return"]:
                callback(exc, exchange, routing_key, msg)
        else:
            raise exc

    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(*args, **kwargs)
        self._method_override = {(60, 50): self._dispatch_basic_return}

    def drain_events(self, timeout=None):
        """Wait for an event on a channel."""
        chanmap = self.channels
        chanid, method_sig, args, content = self._wait_multiple(
                chanmap, None, timeout=timeout)

        channel = chanmap[chanid]

        if content \
        and channel.auto_decode \
        and hasattr(content, 'content_encoding'):
            try:
                content.body = content.body.decode(content.content_encoding)
            except Exception:
                pass

        amqp_method = self._method_override.get(method_sig) or \
                        channel._METHOD_MAP.get(method_sig, None)

        if amqp_method is None:
            raise Exception('Unknown AMQP method (%d, %d)' % method_sig)

        if content is None:
            return amqp_method(channel, args)
        else:
            return amqp_method(channel, args, content)

    def read_timeout(self, timeout=None):
        if timeout is None:
            return self.method_reader.read_method()
        sock = self.transport.sock
        prev = sock.gettimeout()
        if prev != timeout:
            sock.settimeout(timeout)
        try:
            return self.method_reader.read_method()
        finally:
            if prev != timeout:
                sock.settimeout(prev)

    def _wait_multiple(self, channels, allowed_methods, timeout=None):
        for channel_id, channel in channels.iteritems():
            method_queue = channel.method_queue
            for queued_method in method_queue:
                method_sig = queued_method[0]
                if (allowed_methods is None) \
                or (method_sig in allowed_methods) \
                or (method_sig == (20, 40)):
                    method_queue.remove(queued_method)
                    method_sig, args, content = queued_method
                    return channel_id, method_sig, args, content

        # Nothing queued, need to wait for a method from the peer
        read_timeout = self.read_timeout
        wait = self.wait
        while 1:
            channel, method_sig, args, content = read_timeout(timeout)

            if (channel in channels) \
            and ((allowed_methods is None) \
                or (method_sig in allowed_methods) \
                or (method_sig == (20, 40))):
                return channel, method_sig, args, content

            # Not the channel and/or method we were looking for. Queue
            # this method for later
            channels[channel].method_queue.append((method_sig, args, content))

            #
            # If we just queued up a method for channel 0 (the Connection
            # itself) it's probably a close method in reaction to some
            # error, so deal with it right away.
            #
            if channel == 0:
                wait()

### Code from kombu END



import logging
import socket
import errno
import sys
import os
import os.path
import importlib

from amqpservices import amqplib_ext


log = logging.getLogger('amqpservices')

DEFAULT_CONFIG_MODULE_NAME = 'amqpservices_config'


class ConfigData(object):
    """Configuration data processed from the configuration module.
    The original Python module is accessible as `module` attribute.

    Attribute `queues` is dictionary with collected queues configuration
    from all services.

    Sentry object created from specified SENTRY_DSN is accessible
    as `sentry` attribute.
    """

    def __init__(self):
        # Configuration module and queues collected from it
        self._module = None
        self._queues = None
        self._sentry = None

    def _load_config(self):
        if self._module is not None:
            return
        module_from_env = os.getenv('AMQPSERVICES_CONFIG_MODULE')
        if module_from_env:
            log.info('Importing config module from env. variable AMQPSERVICES_CONFIG_MODULE: %s',
                    module_from_env)
            module_name = module_from_env
        else:
            log.info('Importing default config module %s', DEFAULT_CONFIG_MODULE_NAME)
            module_name = DEFAULT_CONFIG_MODULE_NAME
        try:
            self._module = importlib.import_module(module_name)
            self._collect_queues()
        except ImportError:
            msg = 'Cannot import amqpservices config module'
            sys.stderr.write(msg + '\n')
            log.critical(msg)
            raise

    def _collect_queues(self):
        self._queues = {}
        for service in self._module.SERVICES:
            self._queues.update(self._module.SERVICES[service]['queues'])
            self._queues['broadcast.{service}'.format(service=service)] = \
                    {'durable': False, 'want_reply': False}

    @property
    def module(self):
        self._load_config()
        return self._module

    @property
    def queues(self):
        self._load_config()
        return self._queues

    def _create_sentry(self):
        if self._sentry is not None:
            return
        try:
            import raven
            self._sentry = raven.Client(dsn=getattr(config.module, 'SENTRY_DSN', ''))
            log.info('Using sentry for error notification')
        except ImportError:
            class _SentryStub(object):
                def captureMessage(self, msg): pass
                def captureException(self): pass
            self._sentry = _SentryStub()
            log.info('Not using sentry for error notification - no module named "raven"')

    @property
    def sentry(self):
        self._create_sentry()
        return self._sentry


config = ConfigData()


LOCALHOST_CONNECTION = dict(host='localhost', userid='guest', password='guest', ssl=False)

class AMQPBase(object):
    def init_connection_and_channel(self, connparams, connect_timeout):
        log.info('Configured SERVICES=%s', config.module.SERVICES)
        log.info('All queues=%s', config.queues)
        try:
            self.conn = amqplib_ext.Connection(host=connparams['host'], userid=connparams['userid'],
                    password=connparams['password'], ssl=connparams['ssl'],
                        use_threading=False, connect_timeout=connect_timeout)
            self.conn.transport.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.ch = self.conn.channel()
            self.ch.access_request('/data', active=True, write=True)
            self.ch.basic_qos(prefetch_size=0, prefetch_count=1, a_global=False)
        except:
            log.exception('While initializing broker connection')
            sentry.captureException()
            raise
    def _process_events(self, timeout=None):
        while True:
            try:
                self.conn.drain_events(timeout)
                return
            except socket.error, val:
                if val.errno == errno.EINTR:
                    print('Restarting due to interrupting by a signal')
                    pass
                else:
                    raise





import socket

from amqplib.client_0_8 import basic_message

from amqpservices.base import log, config, AMQPBase, LOCALHOST_CONNECTION
from amqpservices.serializers.nop import NopSerializer


class _LazyResult(object):
    '''A lazy result. This class should not be instantiated by a user code.
    '''
    def __init__(self, client, correlation_id, timeout=None):
        self.client = client
        self.correlation_id = correlation_id
        self.timeout = timeout
    def get(self):
        '''Performs the reading of the result and returns it.
        '''
        if self.correlation_id in self.client.waiting_consumed:
            return self.client.waiting_consumed.pop(self.correlation_id)
        return self.client._consume(self.timeout, self.correlation_id, keep_notcorrelated=True)

class Client(AMQPBase):
    undefined_answer = object()
    def __init__(self, connparams=LOCALHOST_CONNECTION, serializer=None):
        self.init_connection_and_channel(connparams, connect_timeout=0.1)
        try:
            self.serializer = serializer or NopSerializer()

            self._correlation_id = 0
            self.waiting_consumed = {}
            self.error_state = False

            self._create_private_reply_queue()
            self._register_service_response_consumer()
        except:
            log.exception('Client.__init__ failed')
            config.sentry.captureException()
            raise

    def _gen_correlation_id(self):
        self._correlation_id += 1
        return str(self._correlation_id)

    def _create_private_reply_queue(self):
        self.reply_queue_name, _, _ = self.ch.queue_declare(exclusive=False)
        self.ch.queue_bind(exchange='amq.direct', queue=self.reply_queue_name, routing_key=self.reply_queue_name)
        log.info('Created reply queue %s' % self.reply_queue_name)

    def _register_service_response_consumer(self):
        def _response_callback(msg):
            log.debug('Got response <%s> <%s>', msg.properties, msg.body)
            self._last_response = msg
            msg.channel.basic_ack(msg.delivery_tag)
        self.ch.basic_consume(queue=self.reply_queue_name, callback=_response_callback)

    def _consume_response(self, timeout=None):
        log.debug('Waiting for response...')
        self._last_response = self.undefined_answer
        # We must be sure that we process a response, not some other AMQP message (eg. flow control msg)
        while self._last_response is self.undefined_answer:
            self._process_events(timeout)
        return self._last_response.properties.get('correlation_id'), self._last_response.body

    def send_request(self, routing_key, request, timeout=None, lazy=False):
        '''Sends a request message to an AMQP broker.

        :param routing_key: an AMQP routing key specifying destination. Queues defined in the config module have routing keys same as their names
        :param request: a body of the request method. In case a serializer is passed in the ``__init__`` it should match it's expected input
        :param timeout: timeout in seconds or None for no timeout. Throws :class:`socket.timeout` in case of timeout.
        :param lazy: returns a :class:`~amqpservices.client._LazyResult` which requires a ``get()`` call to perform actual reading of the result.
        '''

        want_reply = config.queues[routing_key]['want_reply']
        if lazy:
            assert want_reply
        if timeout is None:
            timeout = config.queues[routing_key].get('default_timeout')
        props = {
                'delivery_mode': 2 if config.queues[routing_key]['durable'] else 1,
        }
        if want_reply:
            props['reply_to'] = self.reply_queue_name
            props['correlation_id'] = self._gen_correlation_id()
        msg = basic_message.Message(body=self.serializer.to_bytes(request),
                **props)
        try:
            self.ch.basic_publish(msg, exchange='amq.direct', routing_key=routing_key)
            log.debug('Sent message %s to %s', props, routing_key)
            if lazy:
                return _LazyResult(self, props['correlation_id'])
            if want_reply:
                return self._consume(timeout, props['correlation_id'])
            return None
        except:
            self.error_state = True
            raise

    def _consume(self, timeout, correlation_id, keep_notcorrelated=False):
        while True:
            try:
                resp_correlation_id, res = self._consume_response(timeout=timeout)
            except socket.timeout:
                raise
            else:
                if correlation_id is not None and resp_correlation_id != correlation_id:
                    log.warn('Read message with not expected correlation_id')
                    if keep_notcorrelated:
                        self.waiting_consumed[resp_correlation_id] = res
                        log.warn('waiting_consumed=%s', self.waiting_consumed)
                    continue
                return self.serializer.from_bytes(res)

    def process_broker_events(self):
        '''Communicates with the AMQP broker and possibly processes events.
        It's good to call this method from time to time because a broker can
        asynchronously send events to clients (in case of a web application,
        this should be done after processing an HTTP request).
        '''
        try:
            self._process_events(0)
        except socket.error:
            pass
        else:
            msg = 'Processed AMQP events while process_broker_events()'
            log.warn(msg)
            config.sentry.captureMessage(msg)

    def reset(self):
        '''Resets state of a client by discarding possible responses waiting
        to be consumed and calling :func:`~amqpservices.client.Client.process_broker_events`.
        Raises :class:`~EnvironmentError` in case a connection to a broker is in invalid state.
        '''
        if self.waiting_consumed:
            self.waiting_consumed.clear()
        self.process_broker_events()
        if self.conn.channels is None:
            self.error_state = True
            raise EnvironmentError('Connection to a broker in invalid state')




from .base import config, LOCALHOST_CONNECTION, log
from .client import Client
from .server import Server, worker_id
from .serializers.json import JSONSerializer, JSONDispatcher, expose, JSONCallerProxy, \
    json_server, json_methods_server, json_proxy




class Serializer(object):

    def from_bytes(self, bs):
        '''Creates a rich object from the raw string.
        '''
        raise NotImplementedError()

    def to_bytes(self, p):
        '''Serializes a rich object to a raw string.
        '''
        raise NotImplementedError()


class ResponseProducer(object):

    def __call__(self, request_message):
        raise NotImplementedError()


